/*

the class to build ARG trees from haplotype/genotype data. This is the fast version of Margarita where
we consider the operation starting from the left.


let start with n chromosomes , m markers. 

+ for each node i, we store the nearest neighbor (on the left), the distance, the start and end non-missing locations
the start and end non missing locations help to decrease the running time when compare teo nodes.

if (dist = ed - st) coalescence.

+ we have count[i*4+j] to store number of value 0, 1, 2, and 4 at each marker. 
if (count[i*4] == 1 || count[i*4+1] == 0) && (count[i*4 + 2] == count[i*4+3] == 0) --> Mutation


+ active list: store nodes in active. 







*/






#include <stdio.h>

#include <string.h>
#include <stdlib.h>
#include <time.h>

#include "Const.h"

#include "CNode.h"
#include "CHaplotypes.h"
#include "CARG_f.h"
#include "CQueue.h"
#include "CHap4ARG.h"




#ifndef  __CFARGs__
#define __CFARGs__


#define K10 10000
#define K1000 1000000

#ifndef COAL
#define COAL 1 //coalescent
#endif

#ifndef RECOMB
#define RECOMB 2 //coalescent
#endif



#define DISP 0



//data for one node
//nei : the nodes with maximum overlap from the left
//dist: the overlap distance(from left) 
//st  : the starting location of non '3' or '.'
//ed  : the ending location of non '3' or '.'

//      [st]            [ed]
// .....AAAAAAAAAAAAAAAA.........
//st and ed are used to detect common parts between two nodes
 



struct OnePoint{
  int nei;
  int dist;
  int st;
  int ed;
};


class CFastARG
{
 private:
  CQueue que; 
  //nodes of ARGs
  CNode *nodes[K1000]; //haplotypes
  
  int chr;          //chromosome
  int * marker;     // marker
  int nmarker; //number of marker


  int *count;  //count[i*4 + 0], count[i*4+1], count[i*4+2], count[i*4+3]
  //count[i*4+l]: number of alleles with value l at column i^th

  int nh; //number of haplotypes //
  int nn; //total number of nodes
  int n; //length of nodes

  int BuiltARG;

  /*
    active nodes: nodes are being used to build trees
    inactive nodes: nodes have been used 

    coalesence: (id1, id2) --> id3,     
                 then, id1, id2 -> inactive nodes
		       id3: active nodes

    recombination: id -> (id1, id2)
    then: id -> inactive nodes
          id1, id2: active nodes
   */

  int activenodes[K10]; //store nodes in active. 
  int nactivenodes; //number of active nodes


  //coalescence id1 and id2
  void Coalescence(int id1, int id2);


  // when dealling with chromosome X, we coalescence male samples (haplotype i and i+1) first
  void Coalescence_ChrX(void);

  //count the match between s1 and s2 from left
  //return l: s1[i] ~ s2[i]: i < l; exist j < l: s1[i] != '3' && s2[i] != '3'; ('3': missing or '.')
  int CountLeft(char *s1, char *s2);


  //  count the match between s1 and s2 from left
  //return l: s1[i] ~ s2[i]: i < l; exist j < l: s1[i] != '3' && s2[i] != '3'; ('3': missing or '.')
  //however we know that s1[0]=...=s1[st1-1] = s1[ed1] = ...= s1[nmarker] = '3'
  //                     s2[0]=...=s2[st2-1] = s2[ed2] = ...= s2[nmarker] = '3'
  //so we need to scan only [max(st1, st2) .. min(ed1, ed2)]. 
  //That would save huge running time.
  int CountLeft(char *s1, char *s2, int st1, int ed1, int st2, int ed2);


  //this is a heuristic approach. 
  //we count the distance between nodes[i] and nodes[j] from left
  //without considering missing value, '3' or unphased value '2'
  int ReCountDist(int i, int j);

  //check if s1 and s2 are equal,
  // s1[i] ~ s2[i] && exist j: s1[j] != '3' && s2[j] != '3'
  int Equal(char *s1, char *s2);


  //when we have add a new node or remove an active node, we need to recount "count" vector
  //c = -1: remove, 
  //c =  1: add
  void ReCount(char *s, int c);
  
  //make mutation 
  void Mutation(void);
  //make mutation at site l with n0 values '0' and n1 values '1'
  int Mutation(int l, int n0, int n1);


  //Get ARG trees
  void GetEdges(void);
  void GetEdges(int i);

  //make recombination if node id at location x, x+1
  void RecombinationPoint(int id, int x);

  //make recombination between id1, and id2 at location x and y
  void RecombinationPoint(int id1, int id2,int x, int y);


  //Display all nodes
  void Display(void);

  //Display active nodes/non active node
  void Display(int active);


  //check count vector // expensive
  void CheckCount(void);
  
  //vector of nodes with nearest neighbor, dist, st, and ed (see OnePoint structure for more detail)
  OnePoint *pt;

  //Count nearest neighbor for node id
  void CountNeighbor(int id);


  //remove node id. rmhap = 1: remove from active list. rmhap = 0: do not remove from active list
  void Remove(int id, int rmhap);

  
  //add node id. if add = 1: add to active list. add = 0: do not add to active list when id is already on the active list.
  void Add(int id, int add);

  //initialize pt for node id. 
  void Initpt(int id);

  //re-init pt of node id when adding new node, nid
  void ReInit(int id, int nid);

  
  //Init starting and ed for node id with string s
  void InitSTED(int id, char *s);


  //  CHaplotypes *H; //haplotype or genotyping data


  void PutHap(int id, int k, int v); //put sample id, site k value v
  CARG_f*   Export2CARG_f(CHaplotypes *H);

  int Index[MAXSAMPLE];


 public:
  CFastARG();
  ~CFastARG();
  //read data from file
  void Init(char *fn);


/*
n: number of haplotypes
m: number of markers

S: haplotypes[n][m]: 
S[i][j] = 0, 1, 2 (unphased) or 3 (missing) : i
M: markers: m markers
ichr: chrsomosome
*/
  void Init(int n, int m, char **S, int *M,  int ichr);

  //Get a tree
  void BuildARG(void);
  CARG_f* Build(CHaplotypes* H);
  CARG_f* Build(CHap4ARG *H, char **sNames);

  void InitLikelihood(char ** lknames, int nslk);

  CARG_f*  Export2CARG_f(int chr, int nmarker, int *marker, int nsample, char ** names);
};


#endif
